; persistent relations in ap5
(in-package :ap5)

; In order to make a relation persistent, we log updates.  These are
; occasionally incorporated into a batch save.  On startup the batch
; save and log files are read and the current state recreated in VM.

; We assume that only stored atomic relations are persistent.  
; This implementation does not support maintained relations that are
; derived from persistent relations.  It also does not check constraints
; at startup.  Both these problems could be fixed at the cost of performance.
; Currently I just restore data without rules.  This can be done piecemeal,
; restoring one batch file at a time and one update at a time from log files.
; [Actually, even this can fail in cases where piecemeal updates result in
;  states that cannot be represented.  I suspect that this problem could be
;  solved in practice by logging deletes before inserts.  For now I ignore
;  that problem.]
; The alternative would be to compute one delta that includes all the batch
; and log files, and disable only automation rules.

; Persistent relations participate in the relations
; log-pathname(rel, pathname), batch-pathname(rel, pathname)
; which describe where to save persistent data.

(defrelation log-pathname :types (relation pathname))
(defrelation batch-pathname :types (relation pathname))

; The general solution to the problem of equivalences other than what 
; naturally results from READing the results of PRINT, we provide the
; relation Tuple-printer(rel, fn) which relates a relation to the function
; that is responsible for printing its tuples in such a way that they will
; be restored correctly by READ.
; Fn(tuple, stream) prints tuple (a list of objects) to stream.
; If there is no tuple printer we use print.
(defrelation tuple-printer :types (relation function))

; In order to allow printing code to print one way while saving tuples
; and another for other purposes (readable output to the screen), we
; arrange for saves to be done with the following spec. var. = T.
(defvar *in-tuple-printer* nil)


;; code to log updates


(defvar *logging* t) ; not to be changed in normal operation

(defun log-persistent-updates ()
  (when *logging*
	(map-updates #'log-persistent-update)))

(pushnew 'log-persistent-updates *advice-for-every-transition*)

(defun log-persistent-update (rel tuple tv &rest ignore)
  (declare (ignore ignore))
  (forany (path) s.t. (log-pathname rel path)
	  (with-open-file
	   (log path :if-does-not-exist :create
		:if-exists :append :direction :output)
	   (print (if tv 'add 'delete) log)
	   (let ((*in-tuple-printer* t))
	     (funcall (any x s.t. (tuple-printer rel x) :ifnone #'print)
		      tuple log)))
	  :ifnone nil))


;; code to restore data

(defvar *log-threshold* 20)
; If there are more than this many updates to a relation when we start up,
; we'll batch save it.

; This is not generally needed, but may be useful, e.g., in order to copy
; a DB you might want to copy only the batch files.
(defun save-all (&aux (*logging* nil))
  (loop for (r p) s.t. (batch-pathname r p) do (batch-save r)))

; The programmer must ensure that STARTUP is called before any
; persistent relations are accessed.
(defun startup (&aux (*logging* nil))
  (loop for (rel path) s.t. (batch-pathname rel path) do
	(with-open-file
	 (batch path :direction :input :if-does-not-exist nil)
	 (if batch
	     (let (tuples tup)
	       (loop while (setf tup (read batch nil nil)) do
		     (push tup tuples))
	       (with-dormant-rules t
		  (atomic (loop for tup in tuples do
				(++ apply rel tup)))))
	   (warn "~& file ~A not found" path))))
  (loop for (rel path) s.t. (log-pathname rel path) do
	(with-open-file
	 (log path :direction :input :if-does-not-exist nil)
	 (when log
	     (let (tv tup (n 0))
	       (loop while (setf tv (read log nil nil)) do
		     (incf n)
		     (setf tup (read log)) ; this must not give eof
		     (with-dormant-rules t
			  (if (eq tv 'add)
			      (++ apply rel tup)
			    (-- apply rel tup))))
	       (if (> n *log-threshold*) (batch-save rel)))))))

; The current implementation requires that the relation be totally generable.
; This could be relaxed by restoring a temporary (totally generable) relation
; from the persistent data, then using that to do the save.
(defun batch-save (rel)
  (with-open-file
   (batch (theonly p s.t. (batch-pathname rel p)) :direction :output
	  :if-does-not-exist :create :if-exists :supersede)
   (let ((printer (any x s.t. (tuple-printer rel x) :ifnone #'print))
	 (*in-tuple-printer* t))
     (loop for tup ; if only ... s.t. (apply rel tup) do
	   in (eval (list 'listof rel)) do
	   (funcall printer tup batch))))
  (with-open-file
   (log (theonly p s.t. (log-pathname rel p)) :direction :output
	:if-exists :supersede :if-does-not-exist :create)
   (princ " " log)))


;; printing objects

; For objects that are not normally treated right by PRINT and READ we
; have to arrange for PRINT to produce something that will be READ right.
; I have not attempted to do this automatically for all types.  There are
; some default cases and some utilities to help programmers arrange for the
; right behavior for their own types.
; Here we are concerned mainly with CLOS instances and STRUCTs.
; The default assumption is that all persistent data is saved in the form
; of relations, and that all such objects can be created "empty", the 
; persistent data being filled in incrementally by restoring persistent
; relations.  By default, the only "data" not stored in persistent relations
; is object identity.  Objects are identified by integer IDs which are
; created as needed by reading and writing log and batch save files.

(defvar *highest-id* 0) ; highest ID so far assigned
(defvar *object-table* (make-hash-table)) ; object => (id create-exp)
(defvar *object-id-table* (make-hash-table)) ; id => object

(defun read-obj (stuff &aux (id (car stuff)) (create-exp (cadr stuff)))
  (setf *highest-id* (max id *highest-id*))
  (or (gethash id *object-id-table*)
      (let ((new-obj (eval create-exp)))
	(setf (gethash id *object-id-table*) new-obj)
	(setf (gethash new-obj *object-table*) (list id create-exp))
	new-obj)))

; useful for classes that can be instantiated by just 
; (make-instance '<class-name>)

; try to save original methods

#+clisp
(progn 
 (defvar original-print-object-methods (clos::gf-methods #'print-object))
 (defgeneric old-print-object (obj stream))
 (setf (clos::gf-methods #'old-print-object) original-print-object-methods))
(defmethod print-object ((obj clos::standard-object) stream)
  (if *in-tuple-printer* 
      (let ((lookup (gethash obj *object-table*)))
	(unless lookup
		(setf lookup (list (incf *highest-id*)
				   (list 'make-instance
					 (list 'quote (type-of obj)))))
		(setf (gethash *highest-id* *object-id-table*) obj)
		(setf (gethash obj *object-table*) lookup))
	(princ "#," stream)
	(princ (list 'read-obj (list 'quote lookup)) stream)
	obj)
    (old-print-object obj stream)))

; to be used in print-object methods for struct classes that can
; be instantiated by just (<make-name>)
(defun print-struct (obj stream make-name)
  (let ((lookup (gethash obj *object-table*)))
    (unless lookup
	    (setf lookup (list (incf *highest-id*) (list make-name)))
	    (setf (gethash *highest-id* *object-id-table*) obj)
	    (setf (gethash obj *object-table*) lookup))
    (princ "#," stream)
    (princ (list 'read-obj (list 'quote lookup)) stream)
    obj))

; redefine from ap5

(defun dbobject-print (dbobject stream depth)
  (declare (ignore depth) (special printdbobject))
  ; ~A gets rid of quotes around the strings
  ; The printers should use ~S to preserve packages and case distinctions
  (Cond ;; insert new clause
        (*in-tuple-printer* (print-struct dbobject stream 'make-dbobject))
	((boundp 'printdbobject)
	 (funcall printdbobject dbobject stream))
        ((fboundp 'bootstrap-printer) 
	 (format stream "~A" (bootstrap-printer dbobject)))
	(t (princ '<dbobject> stream))))

#+ignore ;; example
(progn
 (load "~/persist")
 (defclass z () nil)
 (defrelation r1 :types (symbol list))
 (++ log-pathname (relationp 'r1) (parse-namestring "r1.log"))
 (++ batch-pathname (relationp 'r1) (parse-namestring "r1.bat"))
 (defrelation r2 :types (symbol list))
 (++ log-pathname (relationp 'r2) (parse-namestring "r2.log"))
 (++ batch-pathname (relationp 'r2) (parse-namestring "r2.bat"))
 (defun print-r2-tuple (tup stream &aux (*print-circle* t))
   ; save r2 tuples with print-circle
   (print tup stream))
 (++ tuple-printer (relationp 'r2) #'print-r2-tuple)
 (startup)
 ; first time
 (setf z1 (make-instance 'z) z2 (make-instance 'z))
 (setf o1 (make-dbobject) o2 (make-dbobject))
 (setf temp (list 1 2 3))
 (setf temp (cons temp (cdr temp)))
 (atomic (loop for i below 25 do 
    (++ r1 'x (list i z1 o1 temp)) (++ r2 'y (list i z2 o2 temp))))
 ; second time all should be saved (> 20 updates)
)